( electric duet )

include sixtyfive.fs

decimal
: freq ( n -- x ) s>f 12e0 f/ 2e0 fswap f** $100 s>f fswap f/ fround f>s $ff umin ;
hex

data-segment bank new-segment constant song

\ output note directly from osc delay value
: n ( duration delay -- )
  swap t1, t1, ;

: note ( duration note -- )
  over if
    swap t1,
    freq t1,
  else
    2drop
  then ;

#-09 constant note-offset

: c  ( duration octave -- ) #12 *       note-offset + note ;
: c# ( duration octave -- ) #12 * #01 + note-offset + note ; : db c# ;
: d  ( duration octave -- ) #12 * #02 + note-offset + note ;
: d# ( duration octave -- ) #12 * #03 + note-offset + note ; : eb d# ;
: e  ( duration octave -- ) #12 * #04 + note-offset + note ;
: f  ( duration octave -- ) #12 * #05 + note-offset + note ;
: f# ( duration octave -- ) #12 * #06 + note-offset + note ; : gb f# ;
: g  ( duration octave -- ) #12 * #07 + note-offset + note ;
: g# ( duration octave -- ) #12 * #08 + note-offset + note ; : ab g# ;
: a  ( duration octave -- ) #12 * #09 + note-offset + note ;
: a# ( duration octave -- ) #12 * #10 + note-offset + note ; : bb a# ;
: b  ( duration octave -- ) #12 * #11 + note-offset + note ;

: inst ( inst -- )
  $00 t1, t1, ;

: rest ( duration -- )
  dup if
    t1, $00 t1,
  else
    drop
  then ;

: nr ( duration note rest -- )
  -rot note rest ;

struct
  cell% field voice>segment
  cell% field voice>here
  char% field voice>instrument
  char% field voice>duration
  char% field voice>note
end-struct voice%

variable merge.done

\ display voice state
: voice. ( voice -- ) >r cr
  r@ hex.
  r@ voice>here @ hex.
  r@ voice>instrument c@ hex.
  r@ voice>duration c@ hex.
  r@ voice>note c@ hex.
  rdrop ;

\ initialize voice
: voice-init ( voice -- ) >r
  r@ voice>segment @ segment>number @ r@ voice>here !
  0 r@ voice>instrument c!
  0 r@ voice>duration c!
  rdrop ;

\ allocate initialized voice
: new-voice ( -- voice )
  voice% %allot
  data-segment bank new-segment over voice>segment !
  dup voice-init ;

\ set current voice
: switch-voice ( voice -- )
  voice>segment @ switch-segment ;

\ update voice from track data
: voice-next ( voice -- ) >r
  r@ voice>segment @ switch-segment
  begin
    r@ voice>duration c@ 0=
    r@ voice>here @ there u< and
  while
    r@ voice>here @
    dup t1@ dup if
      \ note
      r@ voice>duration c!
      $1 + t1@ r@ voice>note c!
    else
      \ instrument change
      drop
      $1 + t1@ r@ voice>instrument c!
    then
    $2 r@ voice>here +!
  repeat rdrop ;

\ decrement duration of current note
: voice-dec ( duration voice -- ) >r
  r@ voice>duration c@ dup if swap - else nip then
  r@ voice>duration c!
  rdrop ;

\ does this voice have unplayed notes?
: voice-more? ( voice -- f ) >r
  r@ voice>duration c@ 0<>
  r@ voice>here @ r@ switch-voice there u< or
  rdrop ;

\ is a note currently playing?
: voice-active? ( voice -- f )
  voice>note c@ 0<> ;

( merge voice data into electric duet format )

new-voice constant voice0
new-voice constant voice1

variable merge.inst0
variable merge.inst1

\ does at least one voice have an instrument change?
: merge-inst? ( -- f )
  voice0 voice>instrument c@ merge.inst0 @ <>
  voice1 voice>instrument c@ merge.inst1 @ <> or ;

\ compile instrument change command
: merge-inst ( -- )
  $01 t1,
  voice0 voice>instrument c@ dup merge.inst0 ! t1,
  voice1 voice>instrument c@ dup merge.inst1 ! t1, ;

\ compile note command
: merge-note ( -- )
  voice0 voice>duration c@
  voice1 voice>duration c@
  umin dup if
    dup t1,
    negate dup voice0 voice>duration +!
               voice1 voice>duration +!
    voice0 voice>note c@ t1,
    voice1 voice>note c@ t1,
  else
    drop
    true merge.done !
  then ;

: merge-step ( -- )
  voice0 voice-next
  voice1 voice-next
  song switch-segment

  merge-inst? if
    merge-inst
  else
    merge-note
  then ;

: merge ( -- )
  voice0 voice-init
  voice1 voice-init
  false merge.done !
  $ff merge.inst0 !
  $ff merge.inst1 !

  begin
    merge-step
  merge.done @ until ;
 
( arpeggiate multiple voices into one )

variable arp.list 0 arp.list ! \ pointer to zero-terminated list of voice pointers
variable arp.current           \ pointer to list entry of currently active voice
variable arp.dest
variable arp.rate

\ execute xt for all voices in arp.list
: arp-iter ( xt -- )
  arp.list @ begin
  dup @ while
    2dup 2>r
    @ swap execute
    2r> cell+
  repeat 2drop ;

\ display arpeggiator state
: arp. ( -- )
  [ ' voice. ] literal arp-iter ;

\ update all voices from track data
: arp-next ( -- )
  [ ' voice-next ] literal arp-iter ;

: arp-dec-xt ( duration voice -- duration )
  over swap voice-dec ;

\ decrement duration of all current notes
: arp-dec ( duration -- )
  [ ' arp-dec-xt ] literal arp-iter drop ;

\ move arp.current to the next voice, wrapping around if necessary
: arp-round ( -- )
  arp.current @ cell+ dup @ 0= if drop arp.list @ then
  arp.current ! ;

: arp-numactive-xt ( n1 voice -- n2 )
  voice-active? if 1 + then ;

\ how many voices are currently active?
: arp-numactive ( -- n )
  0 [ ' arp-numactive-xt ] literal arp-iter ;

: arp-more-xt ( f1 voice -- f2 )
  voice-more? if true or then ;

\ are there any voices with unplayed notes?
: arp-more? ( -- f )
  false [ ' arp-more-xt ] literal arp-iter ;

: arp-minduration-xt ( n1 voice -- n2 )
  voice>duration c@ dup 0= if drop else umin then ;

\ get the duration of the shortest current note or rest
\ a duration of zero is ignored as this means that the voice has no more data
: arp-minduration ( -- n )
  $100 [ ' arp-minduration-xt ] literal arp-iter ;

\ get the duration of the next arpeggiated note
: arp-duration ( -- n )
  arp-minduration
  arp-numactive 1 > if
    arp.rate @ umin
  then ;

\ move arp.current to the first active voice
\ must be called with at least one voice active
: arp-first-active ( -- )
  begin
    arp.current @ @ voice-active? 0=
  while
    arp-round
  repeat ;

\ update destination instrument if current source instrument differs
: arp-inst ( -- )
  arp.current @ @ voice>instrument c@
  arp.dest @ voice>instrument c@
  <> if
    arp.current @ @ voice>instrument c@
    arp.dest @ 2dup voice>instrument c!
    switch-voice inst
  then ;

\ compile note event to destination voice
: arp-note ( -- )
  arp.dest @ switch-voice
  arp-duration dup t1,
  arp.current @ @ voice>note c@ t1,
  arp-dec ;

\ compile rest event to destination voice
: arp-rest ( -- )
  arp.dest @ switch-voice
  arp-duration dup rest
  arp-dec ;

\ arpeggiate multiple voices
: arp-join ( srclist dest rate -- )
  arp.rate !
  arp.dest !
  dup arp.list !
      arp.current !
  begin
    arp-next
    arp-more?
  while
    arp-numactive if
      arp-first-active
      arp-inst
      arp-note
    else
      arp-rest
    then
    arp-round
  repeat
;

